C $Id: xmeans.f,v 2.0 2004/11/30 09:33:38 tunenori Exp tunenori $
C
C Divide "m" points in "n"-dimensional space into "k" clusters by     
C 	repeating kmeans that divided 2 clusters.
C 	The stopping criterion is BIC.
C Input:
C	real x(m, n)	: "m" points in "n"-dimensional space
C	integer	m, n
C	integer	ik 	: inial number of clusters; the default = 2
C	integer itermax  : max of iterations applied for "kmeans"
C Output:
C	integer	icluster(m) : cluster ID for each element of x
C
C	real centers(m, n) : clusters centers, first 'kx' rows is meaningful
C	real lnL(m)	: log-likelihood, first 'kx' elements is meaningful
C	integer  isize(m) : no. of cluster components, first 'kx' elements is meaningful
C	integer	kx	: final number of generated clusters by "xmeans"
C
C Reference:
C	 Ishioka,T.: ``Extended K-means with an Efficient Estimation of the number of Clusters,'' 
C	Intelligent Data Engineering and Automated Learning --- IDEAL 2000, Second International Conference,
C	Shatin, N.T., Hong Kong, China, December 2000, proceedings 17--22.
C	(Lecture Notes in Computer Science 1983, Kwong Sak Leung, Lai-Wan Chan, Helen Meng (Eds.), Springer, 17--22, 2000)
C
C	http://www.rd.dnc.ac.jp/~tunenori/paperList_e.html
C
	! subroutine xmns(x, m, n, ik, kmax, itermax, icluster, centers, lnL, isize, kx)
	subroutine xmns(centers, lnL, isize, kx)
C
	implicit none

	include "xmeans.h"


	real x(m, n)            ! "m" points in 2 dimensional space
        common /datax/ x

	integer icluster(m)     ! cluster ID for each element of x
        common /icls/ icluster

	real  centers(kmax, n)	! centers(1:kx, n) is meaningful
	real  lnL(kmax)		! lnL(1:kx) is meaningful
	integer  isize(kmax)	! isize(1:kx) is meaningful

	integer	kx	! final number of generated clusters by "xmeans"

	real ctr(ik, n)	! centers, first kmeans uses not k but ik.
	integer  ic1(m) ! cluster ID
        common /icls1/ ic1

	real wss(ik)	! withinss
	integer  nc(ik)	! cluster size

	real yi(m, n)		! yi(1:nc(i), n) is meaningful
	real yi_centers(n)	! c(i, 1:n) by kmeans output 
	real zi_centers(kmax, n) ! At first, same to yi_centers(n), but the row will be added

	integer yi_cluster(m)	! yi_cluster(1:nc(i)) is meaningful

	integer q ! number of clusters; mean and var. for each "n"
	integer	k1, k2	! cluster number

	integer y_continue ! logical
	integer y_subcluster(m) ! kmeans$cluster

	real	y_bic_post(3) ! BIC1, BIC2, BIC
	real 	y_lnL_post(2) ! lnL1, lnL2
	real	y_centers(2, n)
	real 	y_clj1(m, n) ! subcluster-1 data
	real 	y_clj2(m, n) ! subcluster-2 data

	integer y_n1 ! subcluster-1 sample size
	integer y_n2 ! subcluster-2 sample size

	integer lg_bic_prior, lg_lnL_prior ! Logical

	integer	 n_ctrs ! nrow of zi_centers
	integer	 n_lnL  ! nrow of yi_lnL

	real yi_lnL(m)	! yi_lnL(1:n_lnL) is meaningful

	real	y_bic_prior ! bic
	real	y_lnL_prior ! lnL

	real 	stk_data(m, n) ! stacked data
	real 	y2_data(m, n) ! y2 data for stacking
	real 	stk_centers(kmax, n) ! stacked centers
	real 	y2_centers(n)	! y2 center for stacking
	real 	stk_lnL(kmax) ! stacked lnL
	real 	stk_bic(kmax) ! stacked bic
	real 	stk_cls(kmax) ! stacked cls (k2)

	integer	n_stack 	! no. of stacked dataset
	integer	p_stk_data(0:m) ! total number of stacked data; n_stack is meaningful
	integer	p_sum_nc(0:kmax), sum_nc ! sum of nc(i) & pointer; for parsing in margeResult
	integer	p_sum_n_ctrs(0:kmax), sum_n_ctrs ! sum of centers & pointer; for parsing in margeResult
	integer	p_sum_n_lnL(0:kmax), sum_n_lnL ! sum of lnL & pointer; for parsing in margeResult
	integer	p_sum_n_size(0:kmax), sum_n_size ! sum of size & pointer; for parsing in margeResult
	integer	stk_yi_cls(m)
	real	stk_yi_ctrs(kmax, n)
	real	stk_yi_lnL(kmax)
	integer	stk_yi_n_size(kmax)

	integer iy(m)
	integer iy_size	! size of yi
	integer	n_size	! no. of elements of yi_cluster freq. table; given by tabulate

	integer i, j, ii, ip


	
	if (ik .lt. 2) then
		write(*,*) 'Err: xmns, ik =', ik, 'is less than 2'
	endif

	q = 2 * n
	
C	do i = 1, m
C		write(*,*) 'x(', i, ',)=', (x(i, j), j = 1, n)  ! data
C	enddo
	write(*,*) 'm(xmns) =', m
	write(*,*) 'n(xmns) =', n
 	write(*,*) 'kmax(xmns) =', kmax

	call kmeans (x, m, n, m, ctr, ik, itermax, ic1, wss, nc)
	
	sum_nc = 0
	sum_n_ctrs = 0
	sum_n_lnL = 0
	sum_n_size = 0

	do i = 0, m
		p_stk_data(i) = 0
	enddo

	do i = 0, kmax
		p_sum_nc(i) = 0
		p_sum_n_ctrs(i) = 0
		p_sum_n_lnL(i) = 0
		p_sum_n_size(i) = 0
	enddo

	do i = 1, kmax
		do j = 1, n
			stk_yi_ctrs(i, j) = 0
		enddo
		stk_yi_lnL(i) = 0
		stk_yi_n_size(i) = 0
	enddo

	do i = 1, ik	! for each cluster

		! call y_ok(x, m, n, yi, ic1, i, iy_size)	! set yi
		call y_ok(yi, i, iy_size)	! set yi
		
		do j = 1, n
			yi_centers(j) = ctr(i, j)	! set y_centers
			zi_centers(1,j) = yi_centers(j)
		enddo

		write(*,*) 'print yi_centers *********************'
		write(*, *) (yi_centers(j), j = 1, n)

		! write(*,*) 'print yi (i=' , i, ') *********************'
		! do  j = 1, iy_size
		!	write(*, *) (yi(j, ii), ii = 1, n)
		! enddo
		write(*,*) 'print iy_size=', iy_size

		do j = 1, nc(i)
			yi_cluster(j) = 1 ! sub-cluster number should begin from 1
		enddo

		k1 = 1	! cluster number
		k2 = k1 + 1
		lg_bic_prior = 0	! False
		lg_lnL_prior = 0	! False


		n_stack = 0 ! number of stack; divided and unproceeded data are stacked
		p_stk_data(0) = 0

		n_ctrs = 1 ! initial nrow of zi_centers
		n_lnL = 0 ! stored data is unnecessary for lnL

		do

		call split2cls(yi, m, n, iy_size, yi_centers, q, lg_bic_prior, lg_lnL_prior, itermax, y_continue, y_subcluster, y_bic_prior, y_bic_post, y_lnL_prior, y_lnL_post, y_centers, y_clj1, y_clj2, y_n1, y_n2)
		write(*,*) 'y_lnL_post(1) = ', y_lnL_post(1)
		write(*,*) 'y_lnL_post(2) = ', y_lnL_post(2)
		write(*,*) 'y_lnL_prior = ', y_lnL_prior

		if (y_continue .eq. 1) then ! splitting continue
			call updtClusterNum (y_continue, yi_cluster, m, nc(i), k1, k2, y_subcluster)
			! yi_cluster is overwritten

			! write(*,*) 'iy_size = ', iy_size
			! write(*,*) 'y_n1 = ', y_n1
			! write(*,*) 'y_n2 = ', y_n2
			! write(*,*) 'yi_cluster ****************'
			! write(*,*) (yi_cluster(j), j=1, nc(i))
			! write(*,*) 'y_subcluster ****************'
			! write(*,*) (y_subcluster(j), j=1, iy_size)

			call updtCenters (y_continue, zi_centers, kmax, m, n, n_ctrs,  k1, k2, y_centers)
			! zi_centers is overwritten

			n_ctrs = n_ctrs + 1 
			! write(*,*) 'n_lnL = ', n_lnL

			call updtlnL (y_continue, yi_lnL, m, n_lnL, k1, k2, y_lnL_post)
			! write(*,*)(yi_lnL(j), j=1,n_lnL)

			! call print_x2(zi_centers, kmax, n, n_ctrs)

		else	! no node
			if (n_lnL .eq. 0) then
				yi_lnL(1) = y_lnL_prior
				n_lnL = 1
			endif
			if (n_stack .gt. 0) then ! there are stacked data
				write(*,*) 'extract the stacked data(', n_stack, ')...'

				ii = 1
				do ip = p_stk_data(n_stack-1)+1, p_stk_data(n_stack)
					do j = 1, n
						yi(ii, j) = stk_data(ip, j)
					enddo

					! write(*,*) 'stk_data(', ip ,',)=',(yi(ii, j),j=1,2)
					ii = ii + 1
				enddo

				do j = 1, n
					yi_centers(j) = stk_centers(n_stack, j)
				enddo

				y_bic_prior = stk_bic(n_stack)
				y_lnL_prior = stk_lnL(n_stack)
				k1 = stk_cls(n_stack)
				! k2 is unchanged

				iy_size = ii -1
				n_stack = n_stack - 1
				cycle ! next
			endif

C			! no node and no stack		
			write(*,*) 'no node and no stack...'
			exit
		endif
		

C		! splitting continues ... 
C		! y1 = y_clj1 ! sub-cluster-1 data (m x n)
C		! y2 = y_clj2 ! sub-cluster-2 data (m x n)


C		! one-hand repeats recursively

		do ii = 1, y_n1
			do j = 1, n
				yi(ii, j) = y_clj1(ii, j) ! for repeating
			enddo
		enddo

		do ii = 1, y_n2
			do j = 1, n
				y2_data(ii, j) = y_clj2(ii, j) ! for stacking
			enddo
		enddo

		iy_size = y_n1
		! write(*,*) 'iy_size = ', iy_size

		do j = 1, n
			yi_centers(j) = y_centers(1, j) ! for repeating
			y2_centers(j) = y_centers(2, j)! for stacking
		enddo

		y_bic_prior = y_bic_post(1) ! bic
		y_lnL_prior = y_lnL_post(1) ! lnL


C		! other-hand is stacked...


		write(*,*) 'stacking ...'

		n_stack = n_stack + 1 ! total no. of stacked data

		if (n_stack .le. 0) then
			write(*,*) 'Err: xmns() : assertion failed, n_stack =', n_stack
			stop
		endif
		p_stk_data(n_stack) = p_stk_data(n_stack-1) + y_n2 
				! p_stk_data(0) = 0
		! write(*,*) 'p_stk_data(n_stack) =', (p_stk_data(j),j=1,n_stack) 

		ip = p_stk_data(n_stack-1)
		write(*,*) 'ip =', ip
		do ii = 1, y_n2
			do j = 1, n
				stk_data(ip+ii, j) = y2_data(ii, j)
			enddo
		enddo

		! do ii = 1, y_n2
		!	 write(*,*) 'y2_data(',ii,',) =', (y2_data(ii, j),j=1,n)
		! enddo

		do j = 1, n
			stk_centers(n_stack, j) = y2_centers(j)
		enddo

		! write(*,*) 'stk_centers(n_stack, j)', (stk_centers(n_stack, j),j=1,n)
		stk_bic(n_stack) = y_bic_post(2) 
		stk_lnL(n_stack) = y_lnL_post(2) 
		stk_cls(n_stack) = k2

C		! inclement the cluster no.
		k2 = k2 + 1

		enddo ! end of repeat

C		! splitting done
		write(*,*) 'splitting done ...'
		write(*,*) 'main cluster =', i, ' ******************'

		do j = 1, nc(i)
			stk_yi_cls(sum_nc+j) = yi_cluster(j)
		enddo

		sum_nc = sum_nc + nc(i)
		p_sum_nc(i) = sum_nc ! for parsing in margeResult
		! write(*,*) 'p_sum_nc(debug)', p_sum_nc

		do ii = 1, n_ctrs
			do j = 1, n
				stk_yi_ctrs(sum_n_ctrs+ii,j) = zi_centers(ii,j)
			enddo
		enddo

		sum_n_ctrs = sum_n_ctrs + n_ctrs
		p_sum_n_ctrs(i) = sum_n_ctrs ! for parsing in margeResult

		do ii = 1, n_lnL
			stk_yi_lnL(sum_n_lnL+ii) = yi_lnL(ii)
		enddo

		sum_n_lnL = sum_n_lnL + n_lnL
		p_sum_n_lnL(i) = sum_n_lnL ! for parsing in margeResult

		call tabulate (yi_cluster, nc(i), iy, n_size)
		write(*,*) 'n_size', n_size

		do j = 1, n_size
			stk_yi_n_size(sum_n_size+j) = iy(j) ! tabulate
		enddo
		sum_n_size = sum_n_size + n_size
		p_sum_n_size(i) = sum_n_size ! for parsing in margeResult

		! print sub-cluster

		! write(*,*) 'yi_cluster',  (yi_cluster(j), j=1, nc(i))
		! write(*,*) 'zi_centers'
		! call print_x2(zi_centers, kmax, n, n_ctrs)
		! write(*,*) 'yi_lnL', (yi_lnL(j), j=1, n_lnL)
		! write(*,*) 'iy', (iy(j), j=1, n_size)
	enddo

	! write(*,*) 'p_sum_nc', p_sum_nc
	! write(*,*) 'ic1', (ic1(j), j=1, m)
	! write(*,*) 'stk_yi_cls ', stk_yi_cls
	! write(*,*) 'stk_yi_n_size', stk_yi_n_size
	! write(*,*) 'stk_yi_lnL', stk_yi_lnL
	! write(*,*) 'p_sum_n_size', p_sum_n_size

	call mergeResult (stk_yi_cls, stk_yi_n_size,  p_sum_nc, p_sum_n_size, nc)

	if (sum_n_lnL .eq. sum_n_size) then
		kx = sum_n_lnL	! no. of the final clusters 
	else
		write(*,*) 'Err: xmns() : assertion failed, sum_n_lnL =', sum_n_lnL, ', sum_n_size =', sum_n_size
		stop
	endif

	do i = 1, m
		icluster(i) = ic1(i)
	enddo

	do i = 1, sum_n_ctrs
		do j = 1, n
			centers(i,j) = stk_yi_ctrs(i,j)
		enddo
	enddo

	do i = 1, kx
		lnL(i) = stk_yi_lnL(i)
		isize(i) =  stk_yi_n_size(i)
	enddo

	write(*,*) 'final result'
	! write(*,*) 'ic1=', (ic1(i), i=1, m)
	write(*,*) 'centers='
	call print_x2(centers, kmax, n, sum_n_ctrs)
	write(*,*) 'lnL=', (lnL(i),i=1, kx)
	write(*,*) 'isize=', (isize(i), i=1, kx)

	return
	end

C
C Make frequency table. The vector must include positive integer.
C Return the frequency table for the positive integer
C 
	subroutine tabulate (ix, m, iy, imaxx)

	implicit none

	integer m	! (input)
	integer ix(m)	! positive integer (input)
	integer iy(m)	! freq. table assigned for positive ix integer (output)
	integer imaxx	! max of ix elements; iy(1:imaxx) is meaningful (output)

	integer n	! no. of found data
	integer i, j

	n = 0 
	imaxx = 0	! max of ix elements

	write(*,*) 'm(tabulate)=', m
	do j = 1, m
		iy(j) = 0
	enddo

	do j = 1, m
		if (ix(j) .le. 0) then
			n = n + 1
			cycle
		endif
		i = 1 ! initial integer
		do 
			if (ix(j) .eq. i) then
				iy(i) = iy(i) + 1
				n = n + 1
				imaxx = max(i, imaxx)
				exit
			endif
			i = i + 1
		enddo

	enddo
	return
	end

C
C Marge the result of sub-clustering.
C Clutster numbers by first kmeans should be renumbered.
C The other centers, lnLs and sizes are already stacked.
C
	subroutine mergeResult (stk_yi_cls, stk_yi_n_size,  p_sum_nc, p_sum_n_size, nc)

	implicit none
	
	include "xmeans.h"
	integer cl_cluster(m)
	common /icls1/ cl_cluster

	integer stk_yi_cls(m) 
C	real stk_yi_lnL(kmax) 
	integer stk_yi_n_size(kmax)

	integer p_sum_nc(0:kmax)
C	integer p_sum_n_ctrs(0:kmax) ! ctrs are already stacked
C	integer p_sum_n_lnL(0:kmax) ! lnLs are already stacked
	integer p_sum_n_size(0:kmax)

	integer nc(ik) ! the number of sub-cluster for initial cluster
	
	integer kq ! uniq cluster numbers; k should be decremental

C	real centers(ik, m) ! re-assigned centers; (ik, 1:nc(i)) is effective
C	real isize(ik, m) !  re-assigned size; (ik, 1:nc(i)) is effective
C	real lnL(ik, m) ! re-assigned lnL; (ik, 1:nc(i)) is effective
	
	integer xsub(m)
	integer i, i1, i2, j, j1, j2, jj, nn
	integer ik1, kk

	kq = p_sum_n_size(ik)	! cluster number; should be decremental
	kk = kq ! save for initial k

	if (kmax .le. kq) then
		write(*,*) 'Err: mergeResult() : parameter kmax is too small; should be set more than  kq =', kq 
		stop
	endif

	do i = ik, 1, -1 ! roop for main clusters obtained by kmeans
		ik1 = ik - i + 1

		i2 = p_sum_nc(i)
		i1 = p_sum_nc(i-1)+1
		nn = i2 -i1 + 1 ! no. of substituted data
C
C		! roop for subclusters

		j2 = p_sum_n_size(i)
		j1 = p_sum_n_size(i-1)+1
		if (j1 .gt. j2) then
			write(*,*) 'Err: mergeResult() : assertion failed, j1 =', j1, ' j2 =', j2 
			stop
		endif
		if (j1 .le. 0) then
			write(*,*) 'Err: mergeResult() : assertion failed, j1 =', j1
			stop
		endif

		do j = (j2 - j1 + 1), 1,  -1
			do jj = i1, i2
				if (stk_yi_cls(jj) .eq. j) then
					xsub(jj) = kq
				endif
			enddo
			kq = kq - 1
		enddo
		
		! write(*,*) 'xsub', xsub

		jj = i1
		do j = 1, m
			if (cl_cluster(j) .eq. i) then
				cl_cluster(j) = xsub(jj)
				jj = jj +1
			endif
		enddo


	enddo
	if (kq .ne. 0) then
		write(*,*) 'Err: mergeResult() : assertion failed, kq = 0'
		stop
	endif

	return
	end


C
C Update the cluster number by using the result of "split2cls()".
C	v(m) is overwritten by this func-call.
C
	subroutine updtClusterNum (icontinue, v, m, nci, k1, k2, xsub)

	implicit none

	integer icontinue
	integer m ! sample size	
	integer v(m)  ! cluster numbers vector for initial cluster; overwritten
	integer nci ! effective sample size	
	integer k1, k2 ! cluster number should be updated; k1 will become k1 and k2
	integer xsub(m) ! sub-cluster numbers vector of v whose value is k1;
	integer i, j
		! xsub has 1 or 2.
		! partial elements from top are effective.

	if (icontinue .eq. 0) then
		return
	endif

	if (k1 .eq. k2) then
		write(*,*) 'Err: updtClusterNum() : k1 nad k2 shoule be differ'
		stop
	endif
		
	j = 1
	do i = 1, nci
		if (v(i) .eq. k1) then
			if (xsub(j) .eq. 2) then
				v(i) = k2
			else if (xsub(j) .eq. 1) then
				v(i) = k1
			else
				write(*,*) 'Err: updtClustNum(), xsub has illegal value'
			endif
			j = j + 1
		endif
	enddo

	return
	end

C
C Update the cluster centers by using the result of "split2cls()".
C	org_centers is overwritten by this func-call.
C
	subroutine updtCenters (icontinue, org_centers, kmax, m, n, n_ctrs, k1, k2, div_centers)

	implicit none

	integer m, n, k1, k2
	integer kmax
	integer icontinue
	real org_centers(kmax, n) ! original center matrix
	integer n_ctrs		! nrow of org_centers
	real div_centers(2, n) ! divided center matrices, which has 2 rows.
	real z(m, n) ! working area

	integer i, j, ii, nn

	if (n_ctrs .le. 1) then
C		! return div_centers as org_centers	
		do i = 1, 2
			do j = 1, n
				org_centers(i, j) = div_centers(i, j)
			enddo
		enddo
		return
	endif

	if (icontinue .eq. 0) then
		return
	endif

	if (k1 .eq. k2) then
		write(*,*) 'Err: updtCenters() : k1 nad k2 shoule be differ'
		stop
	endif
	
	nn = max(k2, n_ctrs) ! use it when this func is recalled

	ii = 1
	do i = 1, nn
		if (i .eq. k1) then
			do j = 1, n
				z(ii, j) = div_centers(1, j)
			enddo
		else if (i .eq. k2) then
			do j = 1, n
				z(ii, j) = div_centers(2, j)
			enddo
		else
			do j = 1, n
				z(ii, j) = org_centers(i, j)
			enddo
		endif
		ii = ii + 1
	enddo

	do i = 1, (ii-1)
		do j = 1, n
			org_centers(i, j) = z(i, j)
		enddo
	enddo
	return
	end

C
C Update the lnL centers by using the result of "split2cls()".
C	org_lnL is overwritten by this func-call.
C
	subroutine updtlnL (icontinue, org_lnL, m, n_lnL, k1, k2, div_lnL)

	implicit none

	integer icontinue
	integer m
	real org_lnL(m) ! original lnL vector
	integer n_lnL !  elements of org_lnL
	real div_lnL(2) ! divided center matrices, which has 2 rows.
	real z(m) ! original lnL vector
	integer nn, i, j, k1, k2

	if (n_lnL .eq. 0) then
		do i = 1, 2
			org_lnL(i) = div_lnL(i) ! return div_lnL as org_lnL
		enddo
		n_lnL = 2 ! initial setting is 2 as divided lnL
		return
	endif

	if (icontinue .eq. 0) then
		return
	endif

	if (k1 .eq. k2) then
		write(*,*) 'Err: updtCenters() : k1 nad k2 shoule be differ'
		stop
	endif
	
	nn = max(k2, n_lnL) ! use it when this func is recalled

	j = 1
	do i = 1, nn
		if (i .eq. k1) then
			z(j) = div_lnL(1)
		else if (i .eq. k2) then
			z(j) = div_lnL(2)
		else
			z(j) = org_lnL(i)
		endif
		j = j + 1
	enddo

	do i = 1, (j-1)
		org_lnL(i) = z(i)
	enddo

	n_lnL = n_lnL + 1 ! lnL element is added

	return
	end





C
C Split 2 clusters if we would prefer it based on BIC.
	subroutine split2cls(x, m, n, nci, centers, q, lg_bic_prior, lg_lnL_prior, itermax,
     *	 y_continue, y_subcluster, y_bic_prior, y_bic_post,
     *	 y_lnL_prior, y_lnL_post, y_centers, y_clj1, y_clj2, y_n1, y_n2)

	implicit none

	integer m, n
	real x(m, n)
	integer nci ! x (effective) sample size
	real centers(2, n)
	integer q ! a number of parameters
	integer lg_bic_prior ! BIC which x is given; if lg_bic_prior = 0, then we calculate
	integer lg_lnL_prior ! lnL which x is given; if lg_bic_prior = 0, then we calculate

	integer itermax
	integer y_continue ! logical
	integer y_subcluster(m) ! kmeans$cluster

	real	y_bic_prior
	real	y_bic_post(3) ! BIC1, BIC2, BIC
	real	y_lnL_prior
	real 	y_lnL_post(2) ! lnL1, lnL2

	integer k
	parameter (k = 2)	! 2-divide

	real	y_centers(k, n)
	real 	y_clj1(m, n) ! subcluster-1 data
	real 	y_clj2(m, n) ! subcluster-2 data

	integer y_n1 ! subcluster-1 sample size
	integer y_n2 ! subcluster-2 sample size
	! integer	j

	if (lg_bic_prior .eq. 0) then
		call priorBIC(x, m, n, nci, centers, q, y_bic_prior, y_lnL_prior)
		! write(*,*) 'y_bic_prior = ', y_bic_prior
		lg_bic_prior = 1
		lg_lnL_prior = 1
	endif

	call postBICs(x, m, n, nci, q, itermax, y_bic_post, y_lnL_post, y_subcluster, y_centers, y_clj1, y_clj2, y_n1, y_n2)

	! write(*,*) 'y_n1 = ',  y_n1
	! write(*,*) 'y_n2 = ',  y_n2
	! write(*,*) 'y_bic_post(3) = ',  y_bic_post(3)
	! write(*,*) 'y_lnL_post(1) = ',  y_lnL_post(1)
	! write(*,*) 'y_lnL_post(2) = ',  y_lnL_post(2)
	! write(*,*) 'y_subcluster = '
	! write(*,*) (y_subcluster(j), j = 1, nci)
	! write(*,*) 'y_centers(1) = ', (y_centers(1,j), j=1,n)
	! write(*,*) 'y_centers(2) = ', (y_centers(2,j), j=1,n)

	if (y_bic_post(3) .lt. y_bic_prior) then
C		! splitting ...
C		! replace the cluster number to cl$cluster
		y_continue = 1
	else
C		! not splitting ...
C		! return "y_subcluster" stored k
		y_continue = 0
	endif
C	! note that "y_subcluster" gives 1 or 2

	return
	end

C
C Return BIC (prior BIC)
	subroutine priorBIC(x, m, n, nci, centers, q, bic, lnL)

	implicit none

	integer m, n
	real x(m, n)
	integer nci
	real centers(2, n)
	integer q

	real lnL, detVx
	real bic

	! write(*,*) 'nci (in priorBIC) = ', nci
	! write(*,*) 'q (in priorBIC) = ', q

	call calclnL (x, m, n, nci, centers, lnL, detVx)
	bic = -2 * lnL + q * log(real(nci))

	! write(*,*) 'bic (in priorBIC) = ', bic
	! write(*,*) 'lnL (in priorBIC) = ', lnL

	return
	end
C
C Return BICs (two posteror BICs)
	subroutine postBICs(x, m, n, nci, q, itermax, bic, lnL, ic1, ctr, clj1, clj2, y_n1, y_n2)

	implicit none

	integer m, n, nci
	real x(m, n)
	integer q, itermax

	real bic(3)	! BIC1, BIC2, BIC
	real lnL(2)	! lnL1, lnL2

	! integer clsub_cluster(m)	! kmeans$cluster
	integer k
	parameter (k = 2)	! 2-divide
	! real clsub_centers(k, n) ! kmeans$centers

	real clj1(m, n)	! clj1(1:nc(1), n) is effective
	real clj2(m, n) ! clj2(1:nc(2), n) is effective

	integer y_n1, y_n2

	real ctr(k, n)	! centers; kmeans$centers
	real center1(n)	! center of cluster 1
	real center2(n)	! center of cluster 2

	integer ic1(m)	! cluster ID; kmeans$cluster
	real wss(k)     ! withinss
	integer  nc(k)  ! cluster size

	real lnL1, lnL2
	real detVx1, detVx2
	real alpha, beta, r1, r2, s
	real d ! distance
	integer i1, i2, i, j
	real pnorm ! lower prob. of normal distribution

C	! split 2 clusters
	call kmeans (x, m, n, nci, ctr, k, itermax, ic1, wss, nc)

C	! extract sub data
	i1 = 1
	i2 = 1
	do i = 1, nci
		if (ic1(i) .eq. 1) then
			do j = 1, n
				clj1(i1, j) = x(i, j)
			enddo
			i1 = i1 + 1
		else if (ic1(i) .eq. 2) then
			do j = 1, n
				clj2(i2, j) = x(i, j)
			enddo
			i2 = i2 + 1
		else
		endif
	enddo

C	! ratio for pdf
	r1 = nc(1) / (nc(1) + nc(2))	! (0,1)
	r2 = 1.0 - r1			! (0,1)
	y_n1 = nc(1)
	y_n2 = nc(2)

C	! 2 posterior BIC
	do j = 1, n
		center1(j) = ctr(1, j)
		center2(j) = ctr(2, j)
	enddo

	call calclnL (clj1, m, n, nc(1), center1, lnL1, detVx1)
	call calclnL (clj2, m, n, nc(2), center2, lnL2, detVx2)
	
C	! normalizing factor
	if ((detVx1 .ne. detVx1) .or. (detVx2 .ne. detVx2)) then 
C		! if detVx1 or detVx2 is NaN
		beta = 0
	else
		s = 0.0
		do j = 1, n
			d = ctr(1, j) - ctr(2, j)
			s = s + d * d
		enddo
		beta = sqrt( s / (abs(detVx1) + abs(detVx2)) )
	endif

	pnorm = Erf(beta/1.41421354)/2.0 + 0.5 ! pnorm(beta), beta >= 0
	alpha = 0.5 / pnorm

	bic(1) = -2 * lnL1 + q * log(real(nc(1)))
	bic(2) = -2 * lnL2 + q * log(real(nc(2)))
C	! BIC is not (BIC1 + BIC2)
	bic(3) = -2 * lnL1 -2 * lnL2 + 2 * q * log(real(nc(1)+nc(2))) -2 * real(nc(1) + nc(2)) * log(alpha)
	lnL(1) = lnL1
	lnL(2) = lnL2

	return 
	end


C
C Log-likelihood under the assumption of 
C m-dimensional multivariate normal distribution.
	subroutine calclnL(x, m, n, nci, centers, lnL, detVx)

	implicit none

	integer m	! sample size
	integer n	! variable
	real x(m, n)

	integer nci	! effective sample size
	real centers(n) ! centers
	real lnL	! log likelihood
	real detVx	! determinant of Vx
	real VarX(n)	! Var of x
	real invVx(n)	! Inverse of VarX

	integer zero
	data	zero /0/

	real t1, t2, t3
	real xmuij, s
	integer i, j

	do i = 1, n
		if (centers(i) .ne. centers(i)) then ! NaN
			write(*,*) 'Err: Illegal data: centers(1:n) is NaN'
		endif
	enddo

	if (nci .le. 2) then ! few data
		lnL = zero ! treated it as NaN; zero dividing causes SIGABRT
		detVx = zero ! treated it as NaN; zero dividing causes SIGABRT
		return
	endif

	call var(x, m, n, nci, VarX)

	detVx = 1.0
	do i = 1, n
		invVx(i) = 1.0 / VarX(i)
		detVx = detVx * VarX(i)
	enddo
	
	t1 = - n / 2.0 * 1.837877066 ! 1.837... = log(2 * 3.14...)
	t2 = -log(detVx) / 2.0

	s = 0.0
	do i = 1, nci
		do j = 1, n
			xmuij =  x(i, j) - centers(j)
			s = s + xmuij * invVx(j) * xmuij	
		enddo
	enddo
	t3 = -s / 2.0
	lnL = (t1 + t2) * nci + t3

	return
	end

C
C Variance of X
	subroutine var(x, m, n, nci, VarX)

	implicit none

	integer m ! sample size
	integer n ! variable
	real x(m, n)
	integer nci ! effective sample size
	real VarX(n)
	real s(n)
	real ss(n)

	integer i, j
	real xij

	do j = 1, n
		s(j) = 0.0
		ss(j) = 0.0
	enddo

	do i = 1, nci
		do j = 1, n
			xij =  x(i, j)
			s(j) = s(j) + xij
			ss(j) = ss(j) + xij * xij
		enddo
	enddo

	do j = 1, n
		VarX(j) = ( ss(j) - s(j) * s(j) / real(nci) ) / real(nci -1)
	enddo

	return
	end
	
C
C Extract x's whose element's number is "ikey", and set it y.	
C Call side knows y's size in advance.
	! subroutine y_ok(x, m, n, y, key_of_x, ikey, iy_size)
	subroutine y_ok(y, ikey, iy_size)

	implicit none

	include "xmeans.h"
	real x(m, n)
	common /datax/ x
	real y(m, n)

	integer key_of_x(m), ikey
	common /icls1/ key_of_x

	integer iy_size, i, j, ii

	if (m .lt. 1 .or. n .lt. 1) then
		write(*,*) 'Err: Illegal data: m and n must be larger than 1'
		write(*,*) 'm =', m
		write(*,*) 'n =', n
	endif

	ii = 1
	do i = 1, m
		if (key_of_x(i) .eq. ikey) then
			do j = 1, n
				y(ii, j) = x(i, j)
			enddo
			ii = ii + 1
		endif	
	enddo

	iy_size	= ii - 1 ! y's size

	! write(*,*) 'iy_size=' , iy_size
	! write(*,*) 'n=' , n
	! write(*,*) 'print y (ikey=' , ikey, ') *********************'
	!
	! call print_x2(y, m, n, (ii-1))

	return
	end


C
C Print 2 dimensional data;
C Use this for only debug printing.
	subroutine print_x2(x, m, n, im)

	implicit none

	integer m, n
	real x(m, n)
	integer im ! effective m; if im = 0 then we use m instead of this. 
	integer i, j, mm
	
	if (im .gt. 0) then
		mm = im
	else
		mm = m
	endif
	do i = 1, mm
		write(*, *) (x(i, j), j = 1, n)
	enddo
	return
	end
